(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
header {* A Polymorphic State Monad *}
theory State_Monad
imports
  "~~/src/HOL/Library/Monad_Syntax"
begin

text {*The state monad in this theory is very similar to the heap monad in
@{file "~~/src/HOL/Imperative_HOL/Heap_Monad.thy"} but for arbitrary states.
Many definitions and lemmas are just ported from @{text "Heap_Monad"}.*}

subsection {* The Monad *}

subsubsection {*Monad Construction*}
text {*A \emph{state monad} is a transformation function on states, yielding
an additional result. To facilitate the use of @{text partial_function}, the
result is wrappen in an option.*}
datatype ('a, 's) state_monad = State_Monad "'s \<Rightarrow> ('a \<times> 's) option"

text {*To reason about state monads an \emph{existentional view} is provided
via the function @{text "run"}.*}
fun run :: "('a, 's) state_monad \<Rightarrow> 's \<Rightarrow> ('a \<times> 's) option" where
  "run (State_Monad f) = f"

lemma state_monad_cases[case_names succeed fail]:
  fixes f and s
  assumes succeed: "\<And>x s'. run m s = Some (x, s') \<Longrightarrow> P"
      and fail: "run m s = None \<Longrightarrow> P"
  shows "P"
  using assms by (cases "run m s") auto

lemma State_Monad_run[simp]: "State_Monad (run m) = m"
  by (cases m) simp_all

lemma state_monad_eqI: "(\<And>s. run m s = run n s) \<Longrightarrow> m = n"
  by (cases m, cases n) (auto simp: fun_eq_iff)

ML {* structure Run_Simps = Named_Thms(
  val name = "run_simps"
  val description = "simplification rules for running state monads"
) *}
setup Run_Simps.setup

lemma run_Let[run_simps]:
  "run (let x = t in f x) = (let x = t in run (f x))"
  by (simp add: Let_def)


subsubsection {*Specialized Lifters*}

definition state_monad :: "('s \<Rightarrow> 'a \<times> 's) \<Rightarrow> ('a, 's) state_monad" where
  "state_monad f = State_Monad (Some o f)"

lemma run_state_monad[run_simps]: "run (state_monad f) = Some o f"
  by (simp add: state_monad_def)

definition update :: "('s \<Rightarrow> 's) \<Rightarrow> ('s, 's) state_monad " where
  "update f \<equiv> state_monad (\<lambda>s. (s, f s))"

lemma run_update[run_simps]:
  "run (update f) s = Some (s, f s)"
  by (simp add: update_def run_simps)


subsubsection {*Monad Combinators*}

definition return :: "'a \<Rightarrow> ('a, 's) state_monad" where
  "return x = state_monad (Pair x)"

lemma run_return[run_simps]: "run (return x) = Some \<circ> Pair x"
  by (simp add: return_def run_simps)

definition
  bind :: "('a, 's) state_monad \<Rightarrow> ('a \<Rightarrow> ('b, 's) state_monad) \<Rightarrow> ('b, 's) state_monad"
where
  "bind m f = State_Monad (\<lambda>s. case run m s of
      Some (x, s') \<Rightarrow> run (f x) s'
    | None \<Rightarrow> None)"

setup {*
  Adhoc_Overloading.add_variant
    @{const_name Monad_Syntax.bind} @{const_name State_Monad.bind}
*}

lemma run_bind[run_simps]:
  "run m s = Some (x, s') \<Longrightarrow> run (m \<guillemotright>= f) s = run (f x) s'"
  "run m s = None \<Longrightarrow> run (m \<guillemotright>= f) s = None"
  by (auto simp: run_simps bind_def)

lemma update_bind_run[run_simps]: "run (update f \<guillemotright>= g) s = run (g s) (f s)"
  by (simp add: run_simps)

  
subsubsection {*Monad Laws*}

lemma return_bind[simp]: "(return x \<guillemotright>= f) = f x"
  by (rule state_monad_eqI) (simp add: run_simps)

lemma bind_return[simp]: "(m \<guillemotright>= return) = m"
  by (rule state_monad_eqI) (simp split: option.splits add: run_simps bind_def)

lemma bind_assoc[simp]:
  "(((m :: ('a, 's) state_monad) \<guillemotright>= f) \<guillemotright>= g) = (m \<guillemotright>= (\<lambda>x. f x \<guillemotright>= g))"
  by (rule state_monad_eqI) (simp add: bind_def run_simps split: option.splits)


subsubsection {*Successful Computations*}

definition success :: "('a, 's) state_monad \<Rightarrow> 's \<Rightarrow> bool" where
  "success m s \<longleftrightarrow> run m s \<noteq> None"

lemma successE:
  assumes "success m s"
  obtains x s' where "x = fst (the (run n s))"
    and "s' = snd (the (run n s))"
    and "run m s \<noteq> None"
  using assms by (simp add: success_def)

ML {* structure Success_Intros = Named_Thms (
  val name = "success_intros"
  val description = "introduction rules for success of state monads"
) *}
setup Success_Intros.setup

lemma successI: "run m s \<noteq> None \<Longrightarrow> success m s"
  by (simp add: success_def)

lemma success_updateI[success_intros]: "success (update f) s"
  by (rule successI) (simp add: run_simps)

lemma success_state_monadI[success_intros]: "success (state_monad f) s"
  by (rule successI) (simp add: run_simps)

lemma success_LetI[success_intros]:
  "x = t \<Longrightarrow> success (f x) s \<Longrightarrow> success (let x = t in f x) s"
  by (simp add: Let_def)

lemma success_IfI:
  assumes "c \<Longrightarrow> success t s" and "\<not> c \<Longrightarrow> success e s"
  shows "success (if c then t else e) s"
  using assms by (simp add: success_def)


subsubsection {*A Simple Relational Calculus*}

text {*The @{text effect} predicate states that whenever a state monad
@{text m} runs on a state @{text s} and results in a return value @{text x}
and a new state @{text "s'"}, no exception occurs.*}
definition effect :: "('a, 's) state_monad \<Rightarrow> 's \<Rightarrow> 's \<Rightarrow> 'a \<Rightarrow> bool" where
  "effect m s s' x \<longleftrightarrow> run m s = Some (x, s')"

lemma successE': assumes "success m s" obtains s' x where "effect m s s' x"
  using assms by (auto simp: effect_def elim!: successE)

lemma effectI: "run m s = Some (x, s') \<Longrightarrow> effect m s s' x"
  by (simp add: effect_def)

lemma effectE:
  assumes "effect m s s' x"
  obtains "x = fst (the (run m s))"
    and "s' = snd (the (run m s))"
    and "success m s"
proof (rule that)
  from assms have *: "run m s = Some (x, s')" by (simp add: effect_def)
  thus "success m s" by (simp add: success_def)
  from * show "x = fst (the (run m s))" and "s' = snd (the (run m s))" by simp_all
qed

lemma effect_success: "effect m s s' x \<Longrightarrow> success m s"
  by (simp add: effect_def success_def)

lemma success_effect_conv: "success m s \<longleftrightarrow> (\<exists>s' x. effect m s s' x)"
  by (auto elim!: successE' effect_success)

lemma effect_deterministic:
  assumes "effect m s s' x"
    and "effect m s s'' y"
  shows "x = y" and "s' = s''"
  using assms unfolding effect_def by auto

ML {* structure Effect_Intros = Named_Thms (
  val name = "effect_intros"
  val description = "introduction rules for effect on state monads"
) *}
ML {* structure Effect_Elims = Named_Thms (
  val name = "effect_elims"
  val description = "elimination rules for effect on state monads"
) *}
setup "Effect_Intros.setup #> Effect_Elims.setup"

lemma effect_LetI[effect_intros]:
  assumes "x = t" "effect (f x) s s' r"
  shows "effect (let x = t in f x) s s' r"
  using assms by simp

lemma effect_LetE[effect_elims]:
  assumes "effect (let x = t in f x) s s' r"
  obtains "effect (f t) s s' r"
  using assms by simp

lemma effect_IfI[effect_intros]:
  assumes "c \<Longrightarrow> effect t s s' r"
    and "\<not> c \<Longrightarrow> effect e s s' r"
  shows "effect (if c then t else e) s s' r"
  by (cases c) (simp_all add: assms)

lemma effect_IfE[effect_elims]:
  assumes "effect (if c then t else e) s s' r"
  obtains "c" "effect t s s' r"
    | "\<not> c" "effect e s s' r"
  using assms by (cases c) simp_all

lemma effect_updateI[effect_intros]:
  assumes "s' = f s" "x = s"
  shows "effect (update f) s s' x"
  by (rule effectI) (simp add: assms run_simps)

lemma effect_updateE[effect_elims]:
  assumes "effect (update f) s s' x"
  obtains "s' = f s" and "x = s"
  using assms by (rule effectE) (auto simp: run_simps)

lemma effect_state_monadI[effect_intros]:
  assumes "s' = snd (f s)" and "x = fst (f s)"
  shows "effect (state_monad f) s s' x"
  by (rule effectI) (simp add: assms run_simps)

lemma effect_state_monadE[effect_elims]:
  assumes "effect (state_monad f) s s' x"
  obtains "s' = snd (f s)" and "x = fst (f s)"
  using assms by (rule effectE) (simp add: run_simps)

lemma success_returnI[success_intros]: "success (return x) s"
  by (rule successI) (simp add: run_simps)

lemma effect_returnI[effect_intros]: "s = s' \<Longrightarrow> effect (return x) s s' x"
  by (rule effectI) (simp add: run_simps)

lemma effect_returnE[effect_elims]:
  assumes "effect (return x) s s' y"
  obtains "y = x" and "s' = s"
  using assms by (rule effectE) (simp add: run_simps)

lemma run_bind_case:
  "run (m \<guillemotright>= f) s = (case (run m s) of
      Some (x, s') \<Rightarrow> run (f x) s'
    | None \<Rightarrow> None)"
  by (simp add: bind_def)

lemma run_bind_success:
  "success m s \<Longrightarrow> run (m \<guillemotright>= f) s = run (f (fst (the (run m s)))) (snd (the (run m s)))"
  by (cases m s rule: state_monad_cases) (auto elim!: successE simp: bind_def)

lemma success_bind_runI:
  "run m s = Some (x, s') \<Longrightarrow> success (f x) s' \<Longrightarrow> success (m \<guillemotright>= f) s"
  by (auto intro!: successI elim!: successE simp: bind_def)

lemma success_bind_effectI:
  "effect m s s' x \<Longrightarrow> success (f x) s' \<Longrightarrow> success (m \<guillemotright>= f) s"
  by (auto simp: effect_def success_def bind_def)

lemma effect_bindI[effect_intros]:
  assumes "effect m s s' x" and "effect (f x) s' s'' y"
  shows "effect (m \<guillemotright>= f) s s'' y"
  using assms by (auto intro!: effectI elim!: effectE successE) (subst run_bind, simp_all)

lemma effect_bindE[effect_elims]:
  assumes "effect (m \<guillemotright>= f) s s'' y"
  obtains s' x where "effect m s s' x" and "effect (f x) s' s'' y"
  using assms by (auto simp: effect_def bind_def split: option.split_asm)

lemma run_bind_eq_SomeI:
  assumes "run m s = Some (x, s')"
    and "run (f x) s' = Some (y, s'')"
  shows "run (m \<guillemotright>= f) s = Some (y, s'')"
  using assms by (simp add: bind_def)


subsection {* Generic Combinators *}

subsubsection {*Plain Lifting*}

text {*Note the difference to @{text "Heap_Monad.lift"}.*}
fun lift :: "('a \<Rightarrow> 'b) \<Rightarrow> ('a, 's) state_monad \<Rightarrow> ('b, 's) state_monad" where
  "lift f m = m \<guillemotright>= return \<circ> f"


subsubsection {*Folds*}

fun
  foldm :: "('a \<Rightarrow> 'b \<Rightarrow> ('a, 's) state_monad) \<Rightarrow> 'a \<Rightarrow> 'b list \<Rightarrow> ('a, 's) state_monad"
where 
  "foldm f d [] = return d" |
  "foldm f d (x # xs) = do { y \<leftarrow> f d x; foldm f y xs }"

lemma foldm_cong[fundef_cong]:
  assumes "\<And>x y. x \<in> set xs \<Longrightarrow> f y x = g y x"
  shows "foldm f d xs = foldm g d xs"
using assms by (induct xs arbitrary: d) (simp_all)


subsubsection {*Iterators*}

fun
  sequence :: "('a, 's) state_monad list \<Rightarrow> ('a list, 's) state_monad"
where
  "sequence [] = return []" |
  "sequence (m # ms) = do {
    x \<leftarrow> m;
    xs \<leftarrow> sequence ms;
    return (x # xs)
  }"

lemma effect_sequence_NilI[effect_intros]:
  "effect (sequence []) s s []"
  by (simp add: effect_def run_simps)

lemma success_sequence_NilI[success_intros]:
  "success (sequence []) s"
  by (simp add: success_def run_simps)

lemma effect_sequence_ConsI[effect_intros]:
  assumes "effect m s s' x" and "effect (sequence ms) s' s'' xs"
  shows "effect (sequence (m#ms)) s s'' (x#xs)"
  using assms by (auto intro!: effect_intros)

lemma success_sequence_ConsI:
  assumes "success m s" and "success (sequence ms) (snd (the (run m s)))"
  shows "success (sequence (m#ms)) s"
  using assms by (auto elim!: successE intro!: success_intros success_bind_runI)

lemma effect_sequenceI:
  assumes "\<And>i. i < length ms \<Longrightarrow> effect (ms!i) (ss!i) (ss!(Suc i)) (xs!i)"
    and "length xs = length ms" and "length ss = length ms + 1"
  shows "effect (sequence ms) (ss!0) (ss!length ms) xs"
using assms proof (induct ms arbitrary: ss xs)
  case Nil thus ?case by (auto intro!: effect_intros)
next
  case (Cons m ms)
  from `length xs = length (m # ms)`
    obtain y ys where xs: "xs = y # ys" "length ys = length ms" by (induct xs) auto
  from `length ss = length (m # ms) + 1`
    obtain s' ss' where ss: "ss = s'#ss'" "length ss' = length ms + 1" by (induct ss) auto
  have "\<forall>i<length ms. effect (ms!i) (ss'!i) (ss'!Suc i) (ys!i)" unfolding xs ss
  proof (intro allI impI)
    fix i assume "i < length ms"
    hence "Suc i < length (m # ms)" by simp
    with Cons(2)[of "Suc i"] show "effect (ms!i) (ss'!i) (ss'!Suc i) (ys!i)"
      unfolding xs ss by simp
  qed
  hence "effect (sequence ms) (ss'!0) (ss'!length ms) ys"
    using Cons(1) xs ss by simp
  moreover have "effect m s' (ss'!0) y"
  proof -
    have "0 < length (m # ms)" by simp
    moreover have "(m#ms)!0 = m" by simp
    moreover have "ss!0 = s'" by (simp add: ss)
    moreover have "ss!Suc 0 = ss'!0" by (simp add: ss)
    moreover have "xs!0 = y" by (simp add: xs)
    ultimately show ?thesis using Cons(2)[of 0] by simp
  qed
  ultimately show ?case unfolding xs ss by (auto intro!: effect_intros)
qed

(*the premise is stronger than necessary*)
lemma success_sequenceI:
  assumes "\<And>m s. m \<in> set ms \<Longrightarrow> success m s" shows "success (sequence ms) s"
using assms proof (induct ms arbitrary: s)
  case Nil thus ?case by (auto intro!: success_intros)
next
  case (Cons m ms s)
  hence "success m s" by simp
  then obtain x s' where "run m s = Some (x, s')" by (auto elim!: successE)
  moreover from Cons have "success (sequence ms) s'" by simp
  ultimately show "success (sequence (m#ms)) s"
    using success_sequence_ConsI[OF `success m s`, of ms] by simp
qed

definition
  mapM :: "('a \<Rightarrow> ('b, 's) state_monad) \<Rightarrow> 'a list \<Rightarrow> ('b list, 's) state_monad"
where
  "mapM f xs = sequence (map f xs)"

lemma run_mapM[run_simps]: "run (mapM f xs) s = run (sequence (map f xs)) s"
  by (simp add: mapM_def)

lemma success_mapM_NilI[success_intros]: "success (mapM f []) s"
  by (auto intro!: success_intros simp: mapM_def)

lemma success_mapM_ConsI:
  assumes "success (f x) s" and "success (mapM f xs) (snd (the (run (f x) s)))"
  shows "success (mapM f (x#xs)) s"
  using assms and success_sequence_ConsI[of "f x" s "map f xs"]
  by (simp add: mapM_def)

lemma effect_mapM_NilI[effect_intros]: "s = s' \<Longrightarrow> x = [] \<Longrightarrow> effect (mapM f []) s s' x"
  by (auto simp: effect_def run_simps)

lemma effect_mapM_ConsI[effect_intros]:
  assumes "effect (f x) s s' y"
    and "effect (mapM f xs) s' s'' ys"
  shows "effect (mapM f (x#xs)) s s'' (y#ys)"
  using assms by (auto intro!: effect_intros simp: mapM_def)

lemma success_mapMI:
  assumes "\<And>x s. x \<in> set xs \<Longrightarrow> success (f x) s"
  shows "success (mapM f xs) s"
proof -
  from assms have "\<forall>s. \<forall>m \<in> set (map f xs). success m s" by auto
  with success_sequenceI[of "map f xs"] show ?thesis by (simp add: mapM_def)
qed

lemma mapM_cong[fundef_cong]:
  assumes "xs = ys" and "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x"
  shows "mapM f xs = mapM g ys"
proof -
  from assms have "map f xs = map g ys" by simp
  thus ?thesis unfolding mapM_def by simp
qed


subsection {* Partial Function Setup *}

definition state_monad_ord :: "('a, 's) state_monad => ('a, 's) state_monad => bool" where
  "state_monad_ord = img_ord run (fun_ord option_ord)"

definition state_monad_lub :: "(('a,'s) state_monad => bool) => ('a, 's) state_monad" where
  "state_monad_lub = img_lub run State_Monad (fun_lub (flat_lub None))"

interpretation state_monad!: partial_function_definitions state_monad_ord state_monad_lub
proof -
  have "partial_function_definitions (fun_ord option_ord) (fun_lub (flat_lub None))"
    by (rule partial_function_lift) (rule flat_interpretation)
  then have "partial_function_definitions (img_ord run (fun_ord option_ord))
      (img_lub run State_Monad (fun_lub (flat_lub None)))"
    by (rule partial_function_image) (auto intro: state_monad_eqI)
  then show "partial_function_definitions state_monad_ord state_monad_lub"
    by (simp only: state_monad_ord_def state_monad_lub_def)
qed

abbreviation "mono_state_monad \<equiv> monotone (fun_ord state_monad_ord) state_monad_ord"

lemma state_monad_ordI:
  assumes "\<And>s. run m s = None \<or> run m s = run n s"
  shows "state_monad_ord m n"
  using assms unfolding state_monad_ord_def img_ord_def fun_ord_def flat_ord_def
  by blast

lemma state_monad_ordE:
  assumes "state_monad_ord m n"
  obtains "run m s = None" | "run m s = run n s"
  using assms unfolding state_monad_ord_def img_ord_def fun_ord_def flat_ord_def
  by atomize_elim blast

lemma bind_mono[partial_function_mono]:
  fixes B :: "('a \<Rightarrow> ('b, 's) state_monad) \<Rightarrow> ('c, 't) state_monad"
  assumes mf: "mono_state_monad B" and mg: "\<And>y. mono_state_monad (\<lambda>f. C y f)"
  shows "mono_state_monad (\<lambda>f. B f \<guillemotright>= (\<lambda>y. C y f))"
proof (rule monotoneI)
  fix f g :: "'a => ('b, 's) state_monad" assume fg: "fun_ord state_monad_ord f g"
  from mf
  have 1: "state_monad_ord (B f) (B g)" by (rule monotoneD) (rule fg)
  from mg
  have 2: "!!y'. state_monad_ord (C y' f) (C y' g)" by (rule monotoneD) (rule fg)

  have "state_monad_ord (B f \<guillemotright>= (\<lambda>y. C y f)) (B g \<guillemotright>= (\<lambda>y. C y f))"
    (is "state_monad_ord ?L ?R")
  proof (rule state_monad_ordI)
    fix h
    from 1 show "run ?L h = None \<or> run ?L h = run ?R h"
      by (rule state_monad_ordE[where s = h]) (auto simp: run_bind_case)
  qed
  also
  have "state_monad_ord (B g \<guillemotright>= (\<lambda>y'. C y' f)) (B g \<guillemotright>= (\<lambda>y'. C y' g))"
    (is "state_monad_ord ?L ?R")
  proof (rule state_monad_ordI)
    fix h
    show "run ?L h = None \<or> run ?L h = run ?R h"
    proof (cases "run (B g) h")
      case None
      then have "run ?L h = None" by (auto simp: run_bind_case)
      thus ?thesis ..
    next
      case Some
      then obtain r h' where "run (B g) h = Some (r, h')"
        by (metis surjective_pairing)
      then have "run ?L h = run (C r f) h'"
        "run ?R h = run (C r g) h'"
        by (auto simp: run_bind_case)
      with 2[of r] show ?thesis by (auto elim: state_monad_ordE)
    qed
  qed
  finally (state_monad.leq_trans)
  show "state_monad_ord (B f \<guillemotright>= (\<lambda>y. C y f)) (B g \<guillemotright>= (\<lambda>y'. C y' g))" .
qed

end